#load "str.cma";;
loads "Minisat/sat_common_tools.ml";;
loads "Minisat/dimacs_tools.ml";;
loads "Minisat/sat_solvers.ml";;
loads "Minisat/sat_script.ml";;
loads "Minisat/sat_tools.ml";;
loads "Minisat/minisat_parse.ml";;
loads "Minisat/minisat_resolve.ml";;
loads "Minisat/minisat_prove.ml";;
