(* ------------------------------------------------------------------------- *)
(* Library requirements.                                                     *)
(* ------------------------------------------------------------------------- *)

needs "Library/analysis.ml";;
needs "Library/poly.ml";;

(* ------------------------------------------------------------------------- *)
(* The main code.                                                            *)
(* ------------------------------------------------------------------------- *)

loads "Rqe/rqe_lib.ml";;
loads "Rqe/rqe_tactics_ext.ml";;
loads "Rqe/util.ml";;
loads "Rqe/rewrites.ml";;
loads "Rqe/basic.ml";;
loads "Rqe/rqe_num.ml";;
loads "Rqe/rqe_real.ml";;
loads "Rqe/list_rewrites.ml";;
loads "Rqe/rqe_list.ml";;
loads "Rqe/timers.ml";;
loads "Rqe/num_calc_simp.ml";;
loads "Rqe/asym.ml";;
loads "Rqe/rol.ml";;
loads "Rqe/poly_ext.ml";;
loads "Rqe/simplify.ml";;
loads "Rqe/lift_qelim.ml";;
loads "Rqe/defs.ml";;
loads "Rqe/testform_thms.ml";;
loads "Rqe/condense_thms.ml";;
loads "Rqe/inferisign_thms.ml";;
loads "Rqe/matinsert_thms.ml";;
loads "Rqe/signs_thms.ml";;
loads "Rqe/inferpsign_thms.ml";;
loads "Rqe/dedmatrix_thms.ml";;
loads "Rqe/pdivides_thms.ml";;
loads "Rqe/main_thms.ml";;
loads "Rqe/work_thms.ml";;
loads "Rqe/testform.ml";;
loads "Rqe/condense.ml";;
loads "Rqe/inferisign.ml";;
loads "Rqe/matinsert.ml";;
loads "Rqe/signs.ml";;
loads "Rqe/inferpsign.ml";;
loads "Rqe/dedmatrix.ml";;
loads "Rqe/pdivides.ml";;
loads "Rqe/rqe_main.ml";;

(****
loads "Rqe/examples.ml";;
 ****)
